"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const workspace_tools_1 = require("workspace-tools");
const taskId_1 = require("./taskId");
const generateTask_1 = require("./generateTask");
const npmlog_1 = __importDefault(require("npmlog"));
const filterPackages_1 = require("./filterPackages");
const isValidTaskId_1 = require("./isValidTaskId");
function getPipeline(pkg, context) {
    if (context.packagePipelines.has(pkg)) {
        return context.packagePipelines.get(pkg);
    }
    return context.defaultPipeline;
}
/**
 * Gather all the task dependencies defined by the "pipeline" setting, generates a list of edges
 * @param targetTask
 * @param pipeline
 */
function generateTaskDepsGraph(targetTasks, pipeline) {
    const queue = [...targetTasks];
    const visited = new Set();
    const graph = [];
    while (queue.length > 0) {
        const task = queue.shift();
        if (!visited.has(task)) {
            visited.add(task);
            if (Array.isArray(pipeline[task])) {
                if (pipeline[task].length > 0) {
                    for (const depTask of pipeline[task]) {
                        graph.push([depTask, task]);
                        queue.push(depTask);
                    }
                }
                else {
                    graph.push(["", task]);
                }
            }
        }
    }
    return graph;
}
/**
 * A function that creates dependency relationship within the context.tasks and context.taskDepsGraph
 * @param fromTaskId
 * @param toTaskId
 * @param context
 */
function createDep(fromTaskId, toTaskId, context) {
    const { tasks, taskDepsGraph, allPackages } = context;
    if (!isValidTaskId_1.isValidTaskId(fromTaskId, allPackages) ||
        !isValidTaskId_1.isValidTaskId(toTaskId, allPackages)) {
        return;
    }
    taskDepsGraph.push([fromTaskId, toTaskId]);
    if (!tasks.has(fromTaskId)) {
        tasks.set(fromTaskId, () => generateTask_1.generateTask(fromTaskId, context));
    }
    if (!tasks.has(toTaskId)) {
        tasks.set(toTaskId, () => generateTask_1.generateTask(toTaskId, context));
    }
}
/**
 * identify and create a realized task dependency map (discovering)
 *
 * This function will traverse the package dependency graph, and will end up traverse the task depenendencies within the same package (2 layered traversal)
 */
function discoverTaskDeps(context) {
    npmlog_1.default.verbose("discoverTaskDeps", "Discovering task deps");
    const { allPackages, command } = context;
    const filteredPackages = filterPackages_1.filterPackages(context);
    // initialize a queue for a breadth first approach
    const traversalQueue = filteredPackages;
    const visited = new Set();
    const dependentMap = workspace_tools_1.getDependentMap(allPackages);
    while (traversalQueue.length > 0) {
        const pkg = traversalQueue.shift();
        if (!visited.has(pkg)) {
            visited.add(pkg);
            // get pipeline
            const pipeline = getPipeline(pkg, context);
            // establish task graph; push dependents in the traversal queue
            const depTaskGraph = generateTaskDepsGraph(command, pipeline);
            for (const [from, to] of depTaskGraph) {
                const dependentPkgs = dependentMap.get(pkg);
                const toTaskId = taskId_1.getTaskId(pkg, to);
                if (from.startsWith("^") && dependentPkgs !== undefined) {
                    // add task dep from all the package deps within repo
                    for (const depPkg of dependentPkgs) {
                        const fromTaskId = taskId_1.getTaskId(depPkg, from.slice(1));
                        createDep(fromTaskId, toTaskId, context);
                    }
                    // now push the dependents in the traversal queue
                    traversalQueue.push(pkg);
                }
                else {
                    const fromTaskId = taskId_1.getTaskId(pkg, from);
                    // add task dep from same package
                    createDep(fromTaskId, toTaskId, context);
                }
            }
        }
    }
}
exports.discoverTaskDeps = discoverTaskDeps;
